#include <unistd.h>

#include <iostream>
#include <cstdlib>
#include <cstring>

using namespace std;

#include "video.h"

extern int Quittah;

static const bool AnsiOpt             = true;	// Be not brainless?
static const bool Colors              = true;	// Use colors?
static const bool Trust_TabSize       = true;	// Assume regular tabs of 8?
static const bool Use_VerticalTabs    = true;	// Enable VT?
static const bool Trust_DefaultColors = false;	// Trust white on black?

// These aren't ansi and so won't
// pass -ansi without being declared.
//extern "C" int ioctl(int,int,...);
//extern "C" void usleep(unsigned long);

int COLS=80, LINES=25;

Video video;

void Video::ShowCurs(const bool c)
{
	if(isactive)
		cout << (c ? "\33[?25h" : "\33[?25l");
}

void Video::GetScreenGeometry() const
{
#if defined(DJGPP) || defined(__BORLANDC__)
    struct text_info w;
    gettextinfo(&w);
    LINES=w.screenheight;
    COLS =w.screenwidth;
#else
#ifdef TIOCGWINSZ
    struct winsize w;
    if(ioctl(STDOUT_FILENO, TIOCGWINSZ, &w) >= 0)
    {
        LINES=w.ws_row;
        COLS =w.ws_col;
    }
#else
#ifdef WIOCGETD
    struct uwdata w;
    if(ioctl(STDOUT_FILENO, WIOCGETD, &w) >= 0)
    {
        LINES = w.uw_height / w.uw_vs;
        COLS  = w.uw_width / w.uw_hs;
    }
#endif
#endif
#endif
}

void Video::SetAttr(unsigned char n)
{
	if(n==TextAttr)return;
    if(Colors && isactive)
    {
    	cout << "\33[";
    	if(n != 7 || !Trust_DefaultColors)
    	{
   			static const char Swap[] = "04261537";
	    	if(AnsiOpt)
	    	{
	    		bool pp=false, r=false;
    	
		    	if((TextAttr&0x80) > (n&0x80)
    			|| (TextAttr&0x08) > (n&0x08))
		    	{
		    		cout << '0'; pp = true;
		    		TextAttr = 0x07;
		    		r = !Trust_DefaultColors;
	    		}
				
		    	if((n&0x08) && !(TextAttr&0x08)){if(pp)cout<<';';cout<<'1';pp=true;}
		    	if((n&0x80) && !(TextAttr&0x80)){if(pp)cout<<';';cout<<'5';pp=true;}
	       		
		      	if((n&7) != (TextAttr&7) || r)
		      	{
		      		if(pp)cout << ';';
		      		cout << '3' << Swap[n&7];
		      		pp = true;
	    	   	}
		       	if((n&0x70) != (TextAttr&0x70) || r)
		       	{
		       		if(pp)cout << ';';
		       		cout << '4' << Swap[(n>>4)&7];
	       		}
	       	}
   	 	   	else
  		 	{
  		 	   	cout << '0';
  		 	   	if(n&0x08)cout << ";1";
				if(n&0x80)cout << ";5";
				cout << ";4" << Swap[(n>>4)&7]
					 << ";3" << Swap[n&7];
			}
		}
		cout << 'm';
   	}   	
   	TextAttr = n;
}

void Video::GenInit()
{
	mbuf = (unsigned char *)malloc(LINES * COLS
							#ifndef linux
							 * 2
							#endif
							);
	cbuf = (unsigned char *)malloc(LINES * COLS);
	
	Goto(1,1,1,4,3);
}

Video::Video(bool active) : isactive(active)
{
	if(isactive)
		ioctl(0, TCGETA, &back);
	GetScreenGeometry();

	GenInit();

	SetAttr(7);
	clear();
	
	if(isactive)
	{
	#ifndef linux
		cout << "\33)0";	// select line drawing chars as G1
		set = Normal;
	#else
		cout << "\33(U";
	#endif
	}
}
Video::Video(const Video &src) : isactive(false)
{
	GenInit();
	Copy(src);
}

void Video::Copy(const Video &src)
{
	if(isactive)
	{
		unsigned char *m=src.mbuf, *c=src.cbuf;
		#ifdef linux
		unsigned char *M=mbuf, *C=cbuf;
		#endif
	
		for(int y=0; y<LINES; y++)
		{
			#ifndef linux
			Goto(y+1, 1, 0);
			#endif
			for(int x=0; x<COLS; x++)
			{
				#ifdef linux
				register bool temp = (*c==*C && *m==*M);
				C++, M++;
				if(temp)
				{
					c++, m++;
					continue;
				}
				Goto(y+1, x+1, 0);
				#endif
				SetAttr(*c++);
				#ifndef linux
				switch(*m++)
				{
					case 0: NormalSet(); break;
					case 1: SingleSet(); break;
					case 2: DoubleSet(); break;
				}
				#endif
				operator << (*m++);
			}
		}
	}
	else
	{
		memcpy(cbuf, src.cbuf, LINES*COLS);
		memcpy(mbuf, src.mbuf, LINES*COLS
			#ifndef linux
				* 2
			#endif
				);
	}
	
	Goto(src.wx, src.wy, src.wv, src.wv1, src.wv2);	
	SetAttr(src.TextAttr);
}

void Video::Close()
{
	if(!cbuf)return;
	
	if(isactive)
		ioctl(0, TCSETA, &back);	/* restore original terminal mode	*/
	SetAttr(7);
	Goto(LINES, 1);
	if(isactive)
		cout << "\33[?2c"	// normal cursor
	#ifndef linux
			 << "\n\n\xF"	// ^O = select G0
	#else
			 << "\33(B"
	#endif
			 << "\33[?25h";	// show cursor
	ShowCurs(1);
	free(cbuf);
	free(mbuf);
	cbuf = NULL;
}
Video::~Video()
{
	Close();
}
void Video::clear()
{
	if(isactive)
		cout << "\33[2J\33[H";
		
	memset(cbuf, TextAttr, LINES*COLS);
	#ifdef linux
	memset(mbuf, ' ', LINES*COLS);
	#else
	unsigned char *m=mbuf;
	for(int y=0; y<LINES; y++)
		for(int x=0; x<LINES; x++)
		{
			*m++ = (unsigned char)set;
			*m++ = ' ';
		}
	#endif	
	wx=wy=1;
}

#ifndef linux
void Video::SingleSet(void)
{
	if(set==Normal && isactive) cout << (char)('N'&31);
	if(set==Double && isactive) cout << "\33#5";
	set=Single;
}
void Video::DoubleSet(void)
{
	if(set==Normal && isactive) cout << (char)('N'&31);
	set=Double;
}
void Video::NormalSet(void)
{
	if(set!=Normal && isactive) cout << (char)('O'&31);
	set=Normal;
}
#endif

Video &Video::operator << (char c)
{
	if((unsigned char)c >= ' ')
	{
		unsigned d = COLS * (wy-1) + (wx-1);
		cbuf[d] = TextAttr;
#ifdef linux
		mbuf[d] = (unsigned char)c;
#else
		d = COLS*2 * (wy-1) + (wx-1);
		mbuf[d] = (unsigned char)set;
		mbuf[d+1] = (unsigned char)c;
#endif
	}
	
	switch(c)
	{
		case '\n': wx=1, wy++; break;
		case '\b': if(wx>1)wx--; break;
		default:
			#ifndef linux
			switch(c)
			{
				case '': SingleSet(); c=0x30; break;
				case '': SingleSet(); c=0x71; break;
				case '': SingleSet(); c=0x6A; break;
				case '': SingleSet(); c=0x6B; break;
				case '': SingleSet(); c=0x78; break;
				case '': SingleSet(); c=0x61; break;
				
				case '': DoubleSet(); c=0x78; break;
				case '': DoubleSet(); c=0x6D; break;
				case '': DoubleSet(); c=0x6A; break;
				case '': DoubleSet(); c=0x6B; break;
				case '': DoubleSet(); c=0x71; break;
				case '': DoubleSet(); c=0x74; break;
				case '': DoubleSet(); c=0x6C; break;
				case '': DoubleSet(); c=0x6E; break;
				case '': DoubleSet(); c=0x77; break;
				case '': DoubleSet(); c=0x75; break;
				case '': DoubleSet(); c=0x76; break;
				case '': SingleSet(); c=0x7E; break;
				case '': NormalSet(); c='<'; break;
				case '': NormalSet(); c='>'; break;
				
				case '':
				case '':
				case '':
				case '':
					/* ARGH, this creeps my heart */
					SingleSet(); c=0x61; break;
				
				default:
					NormalSet();
			}
			#endif
			wx++;
	}
	if(isactive)
		cout << c;
	return *this;
}	

unsigned char Video::getkey(int wait)
{
	if(!isactive)
	{
		video.Close();
		cerr << "\n\n\ngetkey() from disactive screen!\n";
		exit(-1);
	}
	struct termio term = back;
	term.c_lflag &= ~ECHO & ~ICANON;
	term.c_cc[VMIN] = wait;		/* when set to 0, getch() will return	*/
	ioctl(0, TCSETA, &term);	/* immediately, if no key was pressed	*/
								/* (return value is then EOF)		*/
	unsigned char ch;
	int c = read(0, &ch, 1);
	if(c == 1 || Quittah)return ch;
	Flush();
	usleep(20000);
	return 0;
}
Video &Video::Goto(int y,int x,const int v)
{
	if(v >= 0)
		if(wv != v)
		{
			wv = v;
			ShowCurs(v);
		}
	
	register int oldx=wx, oldy=wy;
	if(x==oldx && y==oldy)goto Done;

	wx=x--, wy=y--;
	oldx--, oldy--;
	
	if(!isactive)goto Done;
	
	if(AnsiOpt)
	{
		if(Trust_TabSize)
		{
			if(x==(oldx&~7)+8)
			{
				cout << '\t';
				goto FixUp;
			}
			if(x==(oldx&~7)+16)
			{
				cout << "\t\t";
				goto FixUp;
			}
		}
		if(!x)
		{
			if(y==oldy+4) { cout << "\n\n\n\n"; goto Done; }
			if(y==oldy+3) { cout << "\n\n\n"; goto Done; }
			if(y==oldy+2) { cout << "\n\n"; goto Done; }
			if(y==oldy+1) { cout << '\n'; goto Done; }
			if(!y) { cout << "\33[H"; goto Done; }
			if(oldx)cout << '\r';
			goto FixUp;
		}
		if(x==oldx)goto FixUp;
		if(x==oldx-1)
		{
			cout << '\b';
		FixUp:	 if(y == oldy-1)cout << "\33[A";
			else if(Use_VerticalTabs && y == oldy+1)cout << '\v';
			else if(Use_VerticalTabs && y == oldy+2)cout << "\v\v";
			else if(Use_VerticalTabs && y == oldy+3)cout << "\v\v\v";
			else if(y < oldy)cout << "\33[" << oldy-y << 'A';
			else if(y > oldy)cout << "\33[" << y-oldy << 'B';
			goto Done;
		}
		if(y == oldy)
		{
			/***/if(x == oldx+1)cout << "\33[C";
			else if(x == oldx-2)cout << "\b\b";
			else if(x == oldx-3)cout << "\b\b\b";
			else if(x < oldx)cout << "\33[" << oldx-x << 'D';
			else if(x > oldx)cout << "\33[" << x-oldx << 'C';
			goto Done;
		}
	}
	cout << "\33[" << wy << ';' << wx << 'H';
	
Done:
	Flush();
	return *this;
}
Video &Video::Goto(const int y, const int x, const int v, int v1,int v2)
{
	if(v1 != wv1)
	{
		if(isactive)
			cout << "\33[?" << (v1 ? 2 : 6) << 'c';
		wv1=v1;
		wv2=v2;
	}
	return Goto(y,x,v);
}
void Video::Flush()
{
	cout << flush;
}
